@@ -1,4 +1,4 @@
-/* cardmgr.h 1.10 1995/04/06 20:02:33 (David Hinds) */
+/* cardmgr.h 1.11 1995/04/20 20:46:16 (David Hinds) */
 
 #define MAX_SOCKS	8
 #define MAX_FUNCTIONS	4
@@ -46,8 +46,19 @@ typedef struct card_info_t {
     struct card_info_t	*next;
 } card_info_t;
 
+typedef struct mtd_ident_t {
+    char		*name;
+    enum {
+	JEDEC_MTD=1, DTYPE_MTD, DEFAULT_MTD
+    } mtd_type;
+    int			dtype, jedec_mfr, jedec_info;
+    char		*module;
+    struct mtd_ident_t	*next;
+} mtd_ident_t;
+    
 extern adjust_list_t	*root_adjust;
 extern device_info_t	*root_device;
 extern card_info_t	*blank_card, *root_card, *root_func;
+extern mtd_ident_t	*root_mtd, *default_mtd;
 
 void parse_configfile(FILE *f);
