@@ -26,13 +26,14 @@
 #include <asm/system.h>
 
 #include "version.h"
-#include "mem_region.h"
 #include "cs_types.h"
+#include "rsrc_mgr.h"
 #include "ss.h"
-/* Throw away prototype for CardServices() */
-#define CardServices XCardServices
+
+#define IN_CARD_SERVICES
 #include "cs.h"
-#undef CardServices
+#include "bulkmem.h"
+#include "cs_internal.h"
 #include "cistpl.h"
 #include "cisreg.h"
 
@@ -41,32 +42,29 @@
 static int handle_apm_event(apm_event_t event);
 #endif
 
-#define CHECK_HANDLE(h) \
-    (((h) == NULL) || ((h)->client_magic != CLIENT_MAGIC))
-
-#define CHECK_SOCKET(s) \
-    (((s) >= sockets) || (socket_table[s]->ss_entry == NULL))
-#define SOCKET(h) (socket_table[(h)->Socket])
-
-#define EVENT(h, e, p) \
-    ((h)->event_handler((e), (p), &(h)->event_callback_args))
-
 #ifdef PCMCIA_DEBUG
-static int pc_debug = PCMCIA_DEBUG;
+int pc_debug = PCMCIA_DEBUG;
 static const char *version =
-    "cs.c 1.92 1995/04/11 18:24:53 (David Hinds)\n";
+    "cs.c 1.98 1995/05/03 16:19:15 (David Hinds)\n";
 #endif
 
 static const char *release =
     "Linux PCMCIA Card Services " CS_RELEASE;
 
+#undef CONFIG_MODVERSIONS
+static struct symbol_table cs_symtab = {
+#include <linux/symtab_begin.h>
+    X(register_ss_entry),
+    X(unregister_ss_entry),
+    X(CardServices),
+    X(MTDHelperEntry),
+#include <linux/symtab_end.h>
+};
+
 /*====================================================================*/
 
 /* Parameters that can be set with 'insmod' */
 
-#ifdef POLL_STATUS
-static int poll_interval	= 100;	/* ticks */
-#endif
 static int setup_delay		= 5;	/* ticks */
 static int shutdown_delay	= 5;	/* ticks */
 static int vcc_settle		= 30;	/* ticks */
@@ -80,9 +78,6 @@ static int cis_speed		= 300;	/* ns */
 /* Access speed for IO windows */
 static int io_speed		= 0;	/* ns */
 
-/* Require strict conformance to CIS rules? */
-static int strict_cis		= 1;
-
 /*====================================================================*/
 
 typedef struct irq_info_t {
@@ -96,90 +91,17 @@ typedef struct irq_info_t {
 
 /*====================================================================*/
 
-#define WINDOW_MAGIC	0xB35C
-
-typedef struct window_t {
-    u_long		magic;
-    int			index;
-    client_handle_t	handle;
-    struct socket_info_t *sock;
-    u_long		base;
-    u_long		num;
-    pcmcia_mem_map ctl;
-} window_t;
-
-/*====================================================================*/
-
-/* Maximum number of memory windows per socket */
-#define MAX_WIN 4
-
-typedef struct socket_info_t {
-    ss_entry_t		ss_entry;
-    u_int		sock;
-    socket_state_t	socket;
-    socket_cap_t	cap;
-    u_long		state;
-    client_handle_t	clients;
-    client_handle_t	reset_handle;
-    struct timer_list	setup, shutdown, unreset;
-    u_long		unreset_timeout;
-    pcmcia_mem_map	cis_mem;
-    u_long		Attributes;
-    u_long		Vcc, Vpp1, Vpp2;
-    u_long		IntType;
-    caddr_t		ConfigBase;
-    u_char		Status, Pin, Copy, Option;
-    u_long		Present;
-    u_long		FirstDevType, FuncCode;
-    u_long		SysInitMask;
-    u_long		ManufCode, ManufInfo;
-    u_long		CardValues;
-    irq_req_t		irq;
-    io_req_t		io;
-    window_t		win[MAX_WIN];
-} socket_info_t;
-
-/* Flags in socket state */
-#define SOCKET_CONFIG_LOCKED	0x01
-#define SOCKET_IRQ_REQ		0x02
-#define SOCKET_IO_REQ		0x04
-#define SOCKET_PRESENT		0x08
-#define SOCKET_SETUP_PENDING	0x10
-#define SOCKET_SHUTDOWN_PENDING	0x20
-#define SOCKET_RESET_PENDING	0x40
-#define SOCKET_SUSPEND		0x80
-#define SOCKET_WIN_REQ(i)	(0x100<<(i))
-
-/*====================================================================*/
-
-#define MAX_SOCK	8
-
 static socket_state_t dead_socket = {
     0, SS_DETECT, 0, 0, 0
 };
 
 /* Table of sockets */
-static socket_t sockets = 0;
-static socket_info_t *socket_table[MAX_SOCK];
-
-typedef struct resource_entry_t {
-    u_long base, num;
-    struct resource_entry_t *next;
-} resource_entry_t;
-
-/* Memory resource database */
-static resource_entry_t mem_db = { 0, 0, &mem_db };
-
-/* IO port resource database */
-static resource_entry_t io_db = { 0, 0, &io_db };
+socket_t sockets = 0;
+socket_info_t *socket_table[MAX_SOCK];
 
 /* Table of IRQ assignments */
 static irq_info_t irq_table[16] = { { 0, 0, 0 }, /* etc */ };
 
-#ifdef POLL_STATUS
-static struct timer_list poll;
-#endif
-
 /*====================================================================*/
 
 /* String tables for error messages */
@@ -227,23 +149,29 @@ static const lookup_t error_table[] = {
 static const lookup_t service_table[] = {
     { AccessConfigurationRegister,	"AccessConfigurationRegister" },
     { AdjustResourceInfo,		"AdjustResourceInfo" },
+    { CheckEraseQueue,			"CheckEraseQueue" },
+    { CloseMemory,			"CloseMemory" },
     { DeregisterClient,			"DeregisterClient" },
     { GetCardServicesInfo,		"GetCardServicesInfo" },
     { GetClientInfo,			"GetClientInfo" },
     { GetConfigurationInfo,		"GetConfigurationInfo" },
     { GetEventMask,			"GetEventMask" },
     { GetFirstClient,			"GetFirstClient" },
+    { GetFirstRegion,			"GetFirstRegion" },
     { GetFirstTuple,			"GetFirstTuple" },
     { GetNextClient,			"GetNextClient" },
+    { GetNextRegion,			"GetNextRegion" },
     { GetNextTuple,			"GetNextTuple" },
     { GetStatus,			"GetStatus" },
     { GetTupleData,			"GetTupleData" },
     { MapMemPage,			"MapMemPage" },
     { ModifyConfiguration,		"ModifyConfiguration" },
     { ModifyWindow,			"ModifyWindow" },
-    { OpenWindow,			"OpenWindow" },
+    { OpenMemory,			"OpenMemory" },
     { ParseTuple,			"ParseTuple" },
+    { ReadMemory,			"ReadMemory" },
     { RegisterClient,			"RegisterClient" },
+    { RegisterMTD,			"RegisterMTD" },
     { ReleaseConfiguration,		"ReleaseConfiguration" },
     { ReleaseIO,			"ReleaseIO" },
     { ReleaseIRQ,			"ReleaseIRQ" },
@@ -256,7 +184,9 @@ static const lookup_t service_table[] = {
     { ResetCard,			"ResetCard" },
     { SetEventMask,			"SetEventMask" },
     { ValidateCIS,			"ValidateCIS" },
+    { WriteMemory,			"WriteMemory" },
     { BindDevice,			"BindDevice" },
+    { BindMTD,				"BindMTD" },
     { ReportError,			"ReportError" },
     { SuspendCard,			"SuspendCard" },
     { ResumeCard,			"ResumeCard" }
@@ -291,6 +221,7 @@ ulong cs_init(u_long kmem_start, u_long kmem_end)
 #ifdef CONFIG_APM
     apm_register_callback(&handle_apm_event);
 #endif
+    register_symtab(&cs_symtab);
     return kmem_start;
 } /* cs_init */
 
@@ -333,132 +264,6 @@ static void busy_loop(u_long len)
     restore_flags(flags);
 } /* busy_loop */
   
-/*======================================================================
-
-    These manage the internal databases of available resources
-    
-======================================================================*/
-
-static int add_interval(resource_entry_t *map, u_long base, u_long num)
-{
-    resource_entry_t *p, *q;
-
-    for (p = map; ; p = p->next) {
-	if ((p != map) && (p->base+p->num-1 >= base))
-	    return CS_IN_USE;
-	if ((p->next == map) || (p->next->base > base+num-1))
-	    break;
-    }
-    q = kmalloc(sizeof(resource_entry_t), GFP_KERNEL);
-    q->base = base; q->num = num;
-    q->next = p->next; p->next = q;
-    return CS_SUCCESS;
-} /* add_interval */
-
-/*====================================================================*/
-
-static int sub_interval(resource_entry_t *map, u_long base, u_long num)
-{
-    resource_entry_t *p, *q;
-
-    for (p = map; ; p = q) {
-	q = p->next;
-	if (q == map)
-	    break;
-	if ((q->base+q->num > base) && (base+num > q->base)) {
-	    if (q->base >= base) {
-		if (q->base+q->num <= base+num) {
-		    /* Delete whole block */
-		    p->next = q->next;
-		    kfree_s(q, sizeof(resource_entry_t));
-		}
-		else {
-		    /* Cut off bit from the front */
-		    q->num = q->base + q->num - base - num;
-		    q->base = base + num;
-		}
-	    }
-	    else if (q->base+q->num <= base+num) {
-		/* Cut off bit from the end */
-		q->num = base - q->base;
-	    }
-	    else {
-		/* Split the block into two pieces */
-		p = kmalloc(sizeof(resource_entry_t), GFP_KERNEL);
-		p->base = base+num;
-		
-		p->num = q->base+q->num - p->base;
-		q->num = base - q->base;
-		p->next = q->next ; q->next = p;
-	    }
-	}
-    }
-    return CS_SUCCESS;
-} /* sub_interval */
-
-/*======================================================================
-
-    These find ranges of I/O ports or memory addresses that are not
-    currently allocated by other devices.
-    
-======================================================================*/
-
-static int find_io_region(ioaddr_t *base, ioaddr_t num, char *name)
-{
-    u_long align;
-    resource_entry_t *m;
-    
-    if (*base != 0) {
-	for (m = io_db.next; m != &io_db; m = m->next) {
-	    if ((*base >= m->base) && (*base+num <= m->base+m->num)) {
-		if (check_region(*base, num))
-		    return -1;
-		else {
-		    request_region(*base, num, name);
-		    return 0;
-		}
-	    }
-	}
-	return -1;
-    }
-    
-    for (align = 1; align < num; align *= 2) ;
-    for (m = io_db.next; m != &io_db; m = m->next) {
-	for (*base = m->base;
-	     *base+align <= m->base + m->num;
-	     *base += align)
-	    if (check_region(*base, num) == 0) {
-		request_region(*base, num, name);
-		return 0;
-	    }
-    }
-    return -1;
-} /* find_io_region */
-
-static int find_mem_region(u_long *base, u_long num, char *name)
-{
-    u_long align;
-    resource_entry_t *m;
-    
-    if (*base != 0) {
-	for (m = mem_db.next; m != &mem_db; m = m->next) {
-	    if ((*base >= m->base) && (*base+num <= m->base+m->num))
-		return register_mem_region(*base, num, name);
-	}
-	return -1;
-    }
-    
-    for (align = 4096; align < num; align *= 2) ;
-    for (m = mem_db.next; m != &mem_db; m = m->next) {
-	for (*base = m->base;
-	     *base+align <= m->base+m->num;
-	     *base += align)
-	    if (register_mem_region(*base, num, name) == 0)
-		return 0;
-    }
-    return -1;
-} /* find_mem_region */
-
 static int try_irq(irq_req_t *req, int irq, int specific)
 {
     irq_info_t *info = &irq_table[irq];
@@ -494,73 +299,6 @@ static int try_irq(irq_req_t *req, int irq, int specific)
     return 0;
 } /* try_irq */
 
-/*======================================================================
-
-    Low-level functions to read and write CIS memory.  I think the
-    write routine is only useful for writing one-byte registers.
-    
-======================================================================*/
-
-static void read_cis_mem(socket_info_t *s, int attr,
-			 long addr, long len, u_char *ptr)
-{
-    pcmcia_mem_map *mem = &s->cis_mem;
-    u_char *sys = (u_char *)(mem->sys_start + (addr & 0x0fff));
-    int inc = 1;
-    
-#ifdef PCMCIA_DEBUG
-    if (pc_debug > 3)
-	printk("cs: read_cis_mem(%d, %#lx, %ld)\n", attr, addr, len);
-#endif
-    mem->card_start = addr & ~0x0fff;
-    mem->flags &= ~MAP_ATTRIB;
-    if (attr) { mem->flags |= MAP_ATTRIB; inc++; }
-    
-    for (; len > 0; mem->card_start += 0x1000, sys -= 0x1000) {
-	s->ss_entry(s->sock, SS_SetMemMap, mem);
-#ifdef PCMCIA_DEBUG
-	if (pc_debug > 3)
-	    printk("cs:  %#2.2x %#2.2x %#2.2x %#2.2x %#2.2x ...\n",
-		   sys[0], sys[inc], sys[2*inc], sys[3*inc], sys[4*inc]);
-#endif
-	for ( ; len > 0; len--, ptr++, sys += inc) {
-	    if (sys > (u_char *)mem->sys_stop) break;
-	    *ptr = *sys;
-	}
-    }
-} /* read_cis_mem */
-
-static void write_cis_mem(socket_info_t *s, int attr,
-			  long addr, long len, u_char *ptr)
-{
-    pcmcia_mem_map *mem = &s->cis_mem;
-    u_char *sys = (u_char *)(mem->sys_start + (addr & 0x0fff));
-    int inc = 1;
-    
-    mem->card_start = addr & ~0x0fff;
-    mem->flags &= ~MAP_ATTRIB;
-    if (attr) { mem->flags |= MAP_ATTRIB; inc++; }
-    
-    for (; len > 0; mem->card_start += 0x1000, sys -= 0x1000) {
-	s->ss_entry(s->sock, SS_SetMemMap, mem);
-	for ( ; len > 0; len--, ptr++, sys += inc) {
-	    if (sys > (u_char *)mem->sys_stop) break;
-	    *sys = *ptr;
-	}
-    }
-} /* write_cis_mem */
-
-static int setup_cis_mem(socket_info_t *s)
-{
-    if (s->cis_mem.sys_start == 0) {
-	if (find_mem_region(&s->cis_mem.sys_start,
-			    0x1000, "card services") != 0)
-	    return CS_OUT_OF_RESOURCE;
-	s->cis_mem.sys_stop = s->cis_mem.sys_start + 0x0fff;
-    }
-    return 0;
-} /* setup_cis_mem */
-
 /*======================================================================
 
     Reset a socket to the default state
@@ -635,6 +373,7 @@ int register_ss_entry(int nsock, ss_entry_t ss_entry)
 	/* base address = 0, map = 0 */
 	s->cis_mem.flags = MAP_ACTIVE;
 	s->cis_mem.speed = cis_speed;
+	s->erase_busy.next = s->erase_busy.prev = &s->erase_busy;
 	
 	for (i = 0; i < sockets; i++)
 	    if (socket_table[i] == NULL) break;
@@ -683,12 +422,23 @@ void unregister_ss_entry(ss_entry_t ss_entry)
 
 ======================================================================*/
 
+static void free_regions(memory_handle_t *list)
+{
+    memory_handle_t tmp;
+    while (*list != NULL) {
+	tmp = *list;
+	*list = tmp->info.next;
+	tmp->region_magic = 0;
+	kfree_s(tmp, sizeof(*tmp));
+    }
+}
+
 static int send_event(socket_info_t *s, event_t event, int priority);
 
 static void shutdown_socket(u_long i)
 {
     socket_info_t *s = socket_table[i];
-
+    
     sti();
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
@@ -700,6 +450,8 @@ static void shutdown_socket(u_long i)
     init_socket(s);
     s->irq.AssignedIRQ = 0;
     s->io.NumPorts1 = s->io.NumPorts2 = 0;
+    free_regions(&s->a_region);
+    free_regions(&s->c_region);
 } /* shutdown_socket */
 
 static void setup_socket(u_long i)
@@ -941,56 +693,6 @@ static int handle_apm_event(apm_event_t event)
 } /* handle_apm_event */
 #endif
 
-/*======================================================================
-
-    Bind_device() associates a device driver with a particular socket.
-    It is normally called by Driver Services after it has identified
-    a newly inserted card.  An instance of that driver will then be
-    eligible to register as a client of this socket.
-    
-======================================================================*/
-
-static int bind_device(bind_req_t *req)
-{
-    client_t *client;
-    socket_info_t *s;
-    ss_callback_t call;
-    int status;
-
-    if (CHECK_SOCKET(req->Socket))
-	return CS_BAD_SOCKET;
-    s = SOCKET(req);
-
-#ifdef MODULE
-    MOD_INC_USE_COUNT;
-#endif
-    if (s->clients == NULL) {
-	call.handler = &parse_events;
-	call.info = s;
-	s->ss_entry(req->Socket, SS_RegisterCallback, &call);
-	s->ss_entry(req->Socket, SS_GetStatus, &status);
-	if (status & SS_DETECT) {
-	    s->state |= SOCKET_SETUP_PENDING;
-	    setup_socket(req->Socket);
-	}
-    }
-    
-    client = (client_t *)kmalloc(sizeof(client_t), GFP_KERNEL);
-    memset(client, '\0', sizeof(client_t));
-    client->client_magic = CLIENT_MAGIC;
-    client->dev_info = req->dev_info;
-    client->Socket = req->Socket;
-    client->state = CLIENT_UNBOUND;
-    client->next = s->clients;
-    s->clients = client;
-#ifdef PCMCIA_DEBUG
-    if (pc_debug)
-	printk("cs: bind_device(): client 0x%p, sock %d, dev %s\n",
-	       client, client->Socket, (char *)client->dev_info);
-#endif
-    return CS_SUCCESS;
-} /* bind_device */
-
 /*======================================================================
 
     Access_configuration_register() reads and writes configuration
@@ -1032,6 +734,97 @@ static int access_configuration_register(client_handle_t handle,
     return CS_SUCCESS;
 } /* access_configuration_register */
 
+/*======================================================================
+
+    Bind_device() associates a device driver with a particular socket.
+    It is normally called by Driver Services after it has identified
+    a newly inserted card.  An instance of that driver will then be
+    eligible to register as a client of this socket.
+    
+======================================================================*/
+
+static int bind_device(bind_req_t *req)
+{
+    client_t *client;
+    socket_info_t *s;
+    ss_callback_t call;
+    int status;
+
+    if (CHECK_SOCKET(req->Socket))
+	return CS_BAD_SOCKET;
+    s = SOCKET(req);
+
+#ifdef MODULE
+    MOD_INC_USE_COUNT;
+#endif
+    if (s->clients == NULL) {
+	call.handler = &parse_events;
+	call.info = s;
+	s->ss_entry(req->Socket, SS_RegisterCallback, &call);
+	s->ss_entry(req->Socket, SS_GetStatus, &status);
+	if (status & SS_DETECT) {
+	    s->state |= SOCKET_SETUP_PENDING;
+	    setup_socket(req->Socket);
+	}
+    }
+    
+    client = (client_t *)kmalloc(sizeof(client_t), GFP_KERNEL);
+    memset(client, '\0', sizeof(client_t));
+    client->client_magic = CLIENT_MAGIC;
+    client->dev_info = req->dev_info;
+    client->Socket = req->Socket;
+    client->state = CLIENT_UNBOUND;
+    client->erase_busy.next = &client->erase_busy;
+    client->erase_busy.prev = &client->erase_busy;
+    client->next = s->clients;
+    s->clients = client;
+#ifdef PCMCIA_DEBUG
+    if (pc_debug)
+	printk("cs: bind_device(): client 0x%p, sock %d, dev %s\n",
+	       client, client->Socket, (char *)client->dev_info);
+#endif
+    return CS_SUCCESS;
+} /* bind_device */
+
+/*======================================================================
+
+    Bind_mtd() associates a device driver with a particular memory
+    region.  It is normally called by Driver Services after it has
+    identified a memory device type.  An instance of the corresponding
+    driver will then be able to register to control this region.
+    
+======================================================================*/
+
+static int bind_mtd(mtd_bind_t *req)
+{
+    socket_info_t *s;
+    memory_handle_t region;
+    
+    if (CHECK_SOCKET(req->Socket))
+	return CS_BAD_SOCKET;
+    s = SOCKET(req);
+
+    if (req->Attributes & REGION_TYPE_AM)
+	region = s->a_region;
+    else
+	region = s->c_region;
+
+    while (region) {
+	if (region->info.CardOffset == req->CardOffset) break;
+	region = region->info.next;
+    }
+    if (!region || (region->mtd != NULL))
+	return CS_BAD_OFFSET;
+    region->dev_info = req->dev_info;
+    
+#ifdef PCMCIA_DEBUG
+    if (pc_debug)
+	printk("cs: bind_mtd(): attr 0x%lx, offset 0x%lx, dev %s\n",
+	       req->Attributes, req->CardOffset, (char *)req->dev_info);
+#endif
+    return CS_SUCCESS;
+} /* bind_mtd */
+
 /*======================================================================
 
     We use standard Linux resource management for ports, address
@@ -1052,10 +845,11 @@ static int adjust_memory(adjust_t *adj)
 
     switch (adj->Action) {
     case ADD_MANAGED_RESOURCE:
-	return add_interval(&mem_db, base, num);
+	if (add_interval(&mem_db, base, num) != 0)
+	    return CS_IN_USE;
 	break;
     case REMOVE_MANAGED_RESOURCE:
-	return sub_interval(&mem_db, base, num);
+	sub_interval(&mem_db, base, num);
 	break;
     default:
 	return CS_UNSUPPORTED_FUNCTION;
@@ -1160,6 +954,7 @@ static int deregister_client(client_handle_t handle)
 {
     client_t **client;
     socket_info_t *s;
+    memory_handle_t region;
     u_long flags;
     int i, sn;
     
@@ -1176,6 +971,15 @@ static int deregister_client(client_handle_t handle)
 	if (handle->state & CLIENT_WIN_REQ(i))
 	    return CS_IN_USE;
 
+    /* Disconnect all MTD links */
+    s = SOCKET(handle);
+    if (handle->mtd_count) {
+	for (region = s->a_region; region; region = region->info.next)
+	    if (region->mtd == handle) region->mtd = NULL;
+	for (region = s->c_region; region; region = region->info.next)
+	    if (region->mtd == handle) region->mtd = NULL;
+    }
+    
     save_flags(flags);
     cli();
     sn = handle->Socket; s = socket_table[sn];
@@ -1289,136 +1093,6 @@ static int get_next_client(client_handle_t *handle, client_req_t *req)
     return CS_SUCCESS;
 } /* get_next_client */
 
-/*======================================================================
-
-    Read tuple codes from attribute memory.
-    
-======================================================================*/
-
-#define LINK_C	0x01
-#define LINK_A	0x02
-#define IS_ATTR 0x04
-
-static int get_next_tuple(client_handle_t handle, tuple_t *tuple);
-
-static int get_first_tuple(client_handle_t handle, tuple_t *tuple)
-{
-    tuple->CISOffset = -1;
-    /* Assume presence of a LONGLINK_C to address 0 */
-    tuple->LinkOffset = 0;
-    tuple->Flags = IS_ATTR | LINK_C;
-    return get_next_tuple(handle, tuple);
-} /* get_first_tuple */
-
-static long follow_link(socket_info_t *s, tuple_t *tuple)
-{
-    u_char link[5];
-    long ofs;
-    
-    switch (tuple->Flags & (LINK_A|LINK_C)) {
-    case LINK_A:
-	ofs = tuple->LinkOffset;
-	tuple->Flags = IS_ATTR;
-	break;
-    case LINK_C:
-	ofs = tuple->LinkOffset;
-	tuple->Flags = 0;
-	break;
-    default:
-	return -1;
-    }
-    read_cis_mem(s, tuple->Flags, ofs, 5, link);
-    if ((link[0] != CISTPL_LINKTARGET) || (link[1] < 3) ||
-	(strncmp(link+2, "CIS", 3) != 0))
-	return -1;
-    return ofs;
-} /* follow_link */
-
-static int get_next_tuple(client_handle_t handle, tuple_t *tuple)
-{
-    socket_info_t *s;
-    u_char link[2], attr;
-    long ofs;
-    int i;
-    
-    if (CHECK_HANDLE(handle))
-	return CS_BAD_HANDLE;
-
-    s = SOCKET(handle);
-    if (setup_cis_mem(s) != 0)
-	return CS_OUT_OF_RESOURCE;
-
-    attr = tuple->Flags & IS_ATTR;
-    if (tuple->CISOffset == -1)
-	ofs = 0;
-    else {
-	if (tuple->TupleLink == 0xff) {
-	    if ((ofs = follow_link(s, tuple)) < 0)
-		return CS_NO_MORE_ITEMS;
-	    attr = tuple->Flags & IS_ATTR;
-	}
-	else
-	    ofs = tuple->CISOffset + (attr ? 2 : 1) * (tuple->TupleLink+2);
-    }
-    
-    for (i = 0; i < 1000; i++) {
-	read_cis_mem(s, attr, ofs, 2, link);
-
-	if (link[0] == CISTPL_NULL) {
-	    ofs += attr ? 2 : 1;
-	    continue;
-	}
-	
-	/* End of chain?  Follow long link if possible */
-	if (link[0] == CISTPL_END) {
-	    if ((ofs = follow_link(s, tuple)) < 0)
-		return CS_NO_MORE_ITEMS;
-	    attr = tuple->Flags & IS_ATTR;
-	    read_cis_mem(s, attr, ofs, 2, link);
-	}
-
-	/* Is this a link tuple?  Make a note of it */
-	if ((link[0] == CISTPL_LONGLINK_A) ||
-	    (link[0] == CISTPL_LONGLINK_C) ||
-	    (link[0] == CISTPL_LINKTARGET) ||
-	    (link[0] == CISTPL_NO_LINK)) {
-	    switch (link[0]) {
-	    case CISTPL_LONGLINK_A:
-		tuple->Flags = attr | LINK_A;
-		read_cis_mem(s, attr, ofs + (attr ? 4 : 2), 4,
-			     (u_char *)&tuple->LinkOffset);
-		break;
-	    case CISTPL_LONGLINK_C:
-		tuple->Flags = attr | LINK_C;
-		read_cis_mem(s, attr, ofs + (attr ? 4 : 2), 4,
-			     (u_char *)&tuple->LinkOffset);
-		break;
-	    case CISTPL_NO_LINK:
-		tuple->Flags = attr;
-		break;
-	    }
-	    if (tuple->Attributes & TUPLE_RETURN_LINK)
-		break;
-	}
-	
-	/* Is this a tuple the user wanted? */
-	if ((tuple->DesiredTuple == RETURN_FIRST_TUPLE) ||
-	    (link[0] == tuple->DesiredTuple))
-	    break;
-	ofs += (attr ? 2 : 1) * (link[1]+2);
-    }
-    if (i == 1000) {
-	printk("cs: overrun in get_next_tuple for socket %d\n",
-	       handle->Socket);
-	return CS_NO_MORE_ITEMS;
-    }
-    
-    tuple->TupleCode = link[0];
-    tuple->TupleLink = link[1];
-    tuple->CISOffset = ofs;
-    return CS_SUCCESS;
-} /* get_next_tuple */
-
 /*====================================================================*/
 
 static int get_status(client_handle_t handle, status_t *status)
@@ -1443,38 +1117,7 @@ static int get_status(client_handle_t handle, status_t *status)
 
 /*====================================================================*/
 
-#define MIN(a,b) (((a)<(b)) ? (a) : (b))
-
-static int get_tuple_data(client_handle_t handle, tuple_t *tuple)
-{
-    socket_info_t *s;
-    u_long len;
-    
-    if (CHECK_HANDLE(handle))
-	return CS_BAD_HANDLE;
-
-    s = SOCKET(handle);
-    if (setup_cis_mem(s) != 0)
-	return CS_OUT_OF_RESOURCE;
-
-    if (tuple->TupleLink < tuple->TupleOffset)
-	return CS_NO_MORE_ITEMS;
-    len = tuple->TupleLink - tuple->TupleOffset;
-    tuple->TupleDataLen = tuple->TupleLink;
-    if (len == 0)
-	return CS_SUCCESS;
-    if (tuple->Flags & IS_ATTR)
-	read_cis_mem(s, 1, tuple->CISOffset+((tuple->TupleOffset+2)<<1),
-		     MIN(len, tuple->TupleDataMax), tuple->TupleData);
-    else
-	read_cis_mem(s, 0, tuple->CISOffset+(tuple->TupleOffset+1),
-		     MIN(len, tuple->TupleDataMax), tuple->TupleData);
-    return CS_SUCCESS;
-} /* get_tuple_data */
-
-/*====================================================================*/
-
-static int map_mem_page(window_t *win, memreq_t *req)
+static int map_mem_page(window_handle_t win, memreq_t *req)
 {
     if ((win == NULL) || (win->magic != WINDOW_MAGIC))
 	return CS_BAD_HANDLE;
@@ -1549,7 +1192,7 @@ static int modify_configuration(client_handle_t handle,
     
 ======================================================================*/
 
-static int modify_window(window_t *win, modwin_t *req)
+static int modify_window(window_handle_t win, modwin_t *req)
 {
     if ((win == NULL) || (win->magic != WINDOW_MAGIC))
 	return CS_BAD_HANDLE;
@@ -1734,15 +1377,14 @@ static int release_irq(client_handle_t handle, irq_req_t *req)
 
 /*====================================================================*/
 
-static int release_window(window_t *win)
+static int release_window(window_handle_t win)
 {
     socket_info_t *s;
     
     if ((win == NULL) || (win->magic != WINDOW_MAGIC))
 	return CS_BAD_HANDLE;
     s = win->sock;
-    if (!(win->handle->state & CLIENT_WIN_REQ(win->index)) ||
-	!(s->state & SOCKET_WIN_REQ(win->index)))
+    if (!(win->handle->state & CLIENT_WIN_REQ(win->index)))
 	return CS_BAD_HANDLE;
 
     /* Shut down memory window */
@@ -1751,7 +1393,7 @@ static int release_window(window_t *win)
     s->state &= ~SOCKET_WIN_REQ(win->index);
 
     /* Release system memory */
-    release_mem_region(win->base, win->num);
+    release_mem_region(win->base, win->size);
     win->handle->state &= ~CLIENT_WIN_REQ(win->index);
 
     win->magic = 0;
@@ -2022,8 +1664,8 @@ static int request_window(client_handle_t *handle, win_req_t *req)
     win->handle = *handle;
     win->sock = s;
     win->base = (u_long)req->Base;
-    win->num = req->Size;
-    if (find_mem_region(&win->base, win->num,
+    win->size = req->Size;
+    if (find_mem_region(&win->base, win->size,
 			(char *)(*handle)->dev_info))
 	return CS_IN_USE;
     req->Base = (caddr_t)win->base;
@@ -2245,39 +1887,6 @@ static int set_event_mask(client_handle_t handle, eventmask_t *mask)
 
 /*====================================================================*/
 
-static int validate_cis(client_handle_t handle, cisinfo_t *info)
-{
-    tuple_t tuple;
-    int ret;
-    
-    if (CHECK_HANDLE(handle))
-	return CS_BAD_HANDLE;
-    
-    info->Chains = 0;
-    tuple.DesiredTuple = RETURN_FIRST_TUPLE;
-    tuple.Attributes = 0;
-    ret = get_first_tuple(handle, &tuple);
-    if (ret != CS_SUCCESS)
-	return CS_SUCCESS;
-
-    /* First tuple should be CISTPL_DEVICE */
-    if (strict_cis && (tuple.TupleCode != CISTPL_DEVICE))
-	return CS_SUCCESS;
-
-    for (info->Chains = 1; info->Chains < 500; info->Chains++) {
-	ret = get_next_tuple(handle, &tuple);
-	if (ret != CS_SUCCESS) break;
-    }
-    if (info->Chains == 500) {
-	printk("cs: too many tuples for socket %d\n", handle->Socket);
-	info->Chains = 0;
-    }
-    
-    return CS_SUCCESS;
-} /* validate_cis */
-
-/*====================================================================*/
-
 static int report_error(char *prefix, void *a1, void *a2)
 {
     int i, func = (int)(a1), ret = (int)(a2);
@@ -2306,7 +1915,7 @@ static int report_error(char *prefix, void *a1, void *a2)
 
 /*====================================================================*/
 
-int CardServices(int func, void *a1, void *a2, void *a3, void *a4)
+int CardServices(int func, void *a1, void *a2, void *a3)
 {
 
 #ifdef PCMCIA_DEBUG
@@ -2327,8 +1936,16 @@ int CardServices(int func, void *a1, void *a2, void *a3, void *a4)
 	return access_configuration_register(a1, a2); break;
     case AdjustResourceInfo:
 	return adjust_resource_info(a1, a2); break;
+    case CheckEraseQueue:
+	return check_erase_queue(a1); break;
+    case CloseMemory:
+	return close_memory(a1); break;
+    case CopyMemory:
+	return copy_memory(a1, a2); break;
     case DeregisterClient:
 	return deregister_client(a1); break;
+    case DeregisterEraseQueue:
+	return deregister_erase_queue(a1); break;
     case GetFirstClient:
 	return get_first_client(a1, a2); break;
     case GetCardServicesInfo:
@@ -2337,8 +1954,12 @@ int CardServices(int func, void *a1, void *a2, void *a3, void *a4)
 	return get_configuration_info(a1, a2); break;
     case GetNextClient:
 	return get_next_client(a1, a2); break;
+    case GetFirstRegion:
+	return get_first_region(a1, a2); break;
     case GetFirstTuple:
 	return get_first_tuple(a1, a2); break;
+    case GetNextRegion:
+	return get_next_region(a1, a2); break;
     case GetNextTuple:
 	return get_next_tuple(a1, a2); break;
     case GetStatus:
@@ -2351,10 +1972,18 @@ int CardServices(int func, void *a1, void *a2, void *a3, void *a4)
 	return modify_configuration(a1, a2); break;
     case ModifyWindow:
 	return modify_window(a1, a2); break;
+    case OpenMemory:
+	return open_memory(a1, a2);
     case ParseTuple:
 	return parse_tuple(a1, a2, a3); break;
+    case ReadMemory:
+	return read_memory(a1, a2, a3); break;
     case RegisterClient:
 	return register_client(a1, a2); break;
+    case RegisterEraseQueue:
+	return register_erase_queue(a1, a2); break;
+    case RegisterMTD:
+	return register_mtd(a1, a2); break;
     case ReleaseConfiguration:
 	return release_configuration(a1, a2); break;
     case ReleaseIO:
@@ -2377,8 +2006,12 @@ int CardServices(int func, void *a1, void *a2, void *a3, void *a4)
 	return set_event_mask(a1, a2); break;
     case ValidateCIS:
 	return validate_cis(a1, a2); break;
+    case WriteMemory:
+	return write_memory(a1, a2, a3); break;
     case BindDevice:
 	return bind_device(a1); break;
+    case BindMTD:
+	return bind_mtd(a1); break;
     case ReportError:
 	return report_error(a1, a2, a3); break;
     case SuspendCard:
@@ -2391,7 +2024,7 @@ int CardServices(int func, void *a1, void *a2, void *a3, void *a4)
 	return insert_card(a1, a2); break;
     default:
 	return CS_UNSUPPORTED_FUNCTION; break;
-    };
+    }
     
 } /* CardServices */
 
