@@ -54,12 +54,14 @@
 #define DE650_START_PG	0x40	/* First page of TX buffer */
 #define DE650_STOP_PG	0x80	/* Last page +1 of RX ring */
 
+#define DE650_RDC_TIMEOUT 0x02	/* Max wait in jiffies for Tx RDC */
+
 static char *if_names[] = {"10baseT", "undefined", "undefined", "BNC"};
 
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-    "de650_cs.c 1.62 1995/04/13 04:52:38 (David Hinds)\n";
+    "de650_cs.c 1.65 1995/05/15 04:46:01 (David Hinds)\n";
 #endif
 
 /*====================================================================*/
@@ -324,6 +326,7 @@ static void de650_config(dev_link_t *link)
 #endif
 
     do {
+	tuple.Attributes = 0;
 	tuple.DesiredTuple = CISTPL_CONFIG;
 	i = CardServices(GetFirstTuple, handle, &tuple);
 	if (i != CS_SUCCESS) break;
@@ -667,14 +670,14 @@ static int de650_block_input(struct device *dev, int count,
     if (ei_debug > 4)
 	if (count != 4) printk("%s: [bi=%d]\n", dev->name, count+4);
 #endif
-    if (ei_status.dmaing) {
+    if (set_bit(0,(void *)&ei_status.dmaing)) {
 	if (ei_debug > 0)
 	    printk("%s: DMAing conflict in de650_block_input."
 		   "[DMAstat:%1x][irqlock:%1x]\n",
 		   dev->name, ei_status.dmaing, ei_status.irqlock);
 	return 0;
     }
-    ei_status.dmaing |= 0x01;
+    ei_status.dmaing |= 0x02;
     outb_p(E8390_NODMA+E8390_PAGE0+E8390_START, DE650_BASE+DE650_CMD);
     outb_p(count & 0xff, DE650_BASE + EN0_RCNTLO);
     outb_p(count >> 8, DE650_BASE + EN0_RCNTHI);
@@ -706,7 +709,8 @@ static int de650_block_input(struct device *dev, int count,
 		   dev->name, ring_offset + xfer_count, addr);
     }
 #endif
-    ei_status.dmaing &= ~0x01;
+    outb_p(ENISR_RDC, DE650_BASE + EN0_ISR);
+    ei_status.dmaing &= ~0x03;
     return ring_offset + count;
 } /* de650_block_input */
   
@@ -719,7 +723,8 @@ static void de650_block_output(struct device *dev, int count,
 #ifdef PCMCIA_DEBUG
     int retries = 0;
 #endif
-
+    u_long dma_start;
+    
 #ifdef PCMCIA_DEBUG
     if (ei_debug > 4) printk("%s: [bo=%d]\n", dev->name, count);
 #endif
@@ -728,20 +733,23 @@ static void de650_block_output(struct device *dev, int count,
        I should check someday. */
     if (count & 0x01)
 	count++;
-    if (ei_status.dmaing) {
+    if (set_bit(0, (void *)&ei_status.dmaing)) {
 	if (ei_debug > 0) 
 	    printk("%s: DMAing conflict in de650_block_output."
 		   "[DMAstat:%1x][irqlock:%1x]\n",
 		   dev->name, ei_status.dmaing, ei_status.irqlock);
 	  return;
     }
-    ei_status.dmaing |= 0x02;
+    ei_status.dmaing |= 0x04;
     /* We should already be in page 0, but to be safe... */
     outb_p(E8390_PAGE0+E8390_START+E8390_NODMA, DE650_BASE+DE650_CMD);
 
 #ifdef PCMCIA_DEBUG
  retry:
 #endif
+
+    outb_p(ENISR_RDC, DE650_BASE + EN0_ISR);
+    
     /* Now the normal output. */
     outb_p(count & 0xff, DE650_BASE + EN0_RCNTLO);
     outb_p(count >> 8,   DE650_BASE + EN0_RCNTHI);
@@ -750,10 +758,12 @@ static void de650_block_output(struct device *dev, int count,
     
     outb_p(E8390_RWRITE+E8390_START, DE650_BASE+DE650_CMD);
     outsw(DE650_BASE + DE650_DATAPORT, buf, count>>1);
+
+    dma_start = jiffies;
     
+#ifdef PCMCIA_DEBUG
     /* This was for the ALPHA version only, but enough people have
        encountering problems that it is still here. */
-#ifdef PCMCIA_DEBUG
     if (ei_debug > 4) {	/* DMA termination address check... */
 	int addr, tries = 20;
 	do {
@@ -774,10 +784,19 @@ static void de650_block_output(struct device *dev, int count,
 	}
     }
 #endif
+
+    while ((inb_p(DE650_BASE + EN0_ISR) & ENISR_RDC) == 0)
+	if (jiffies - dma_start > DE650_RDC_TIMEOUT) {
+	    printk("%s: timeout waiting for Tx RDC.\n", dev->name);
+	    de650_reset_8390(dev);
+	    NS8390_init(dev, 1);
+	    break;
+	}
+    
     if (delay_output)
 	udelay((long)delay_time);
-    ei_status.dmaing &= ~0x02;
-    return;
+    ei_status.dmaing &= ~0x05;
+
 } /* de650_block_output */
 
 /*====================================================================*/
@@ -794,7 +813,7 @@ int init_module(void)
 #endif
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("de650_cs: Card Services release does not match!");
+	printk("de650_cs: Card Services release does not match!\n");
 	return -1;
     }
     register_pcmcia_driver(&dev_info, &de650_attach, &de650_detach);
