@@ -28,13 +28,14 @@
 #include "version.h"
 #include "cs_types.h"
 #include "cs.h"
+#include "bulkmem.h"
 #include "cistpl.h"
 #include "ds.h"
 
 #ifdef PCMCIA_DEBUG
 int pc_debug = PCMCIA_DEBUG;
 static const char *version =
-    "ds.c v1.41: 1995/04/11 18:25:31 (David Hinds)\n";
+    "ds.c v1.43: 1995/05/15 01:43:40 (David Hinds)\n";
 #endif
 
 /*====================================================================*/
@@ -360,6 +361,39 @@ static int ds_event(event_t event, int priority,
     return 0;
 } /* ds_event */
 
+/*======================================================================
+
+    bind_mtd() connects a memory region with an MTD client.
+    
+======================================================================*/
+
+static int bind_mtd(int i, mtd_info_t *mtd_info)
+{
+    struct driver_info_t *driver;
+    mtd_bind_t bind_req;
+    int ret;
+
+    for (driver = root_driver; driver; driver = driver->next)
+	if (strcmp((char *)driver->dev_info,
+		   (char *)mtd_info->dev_info) == 0)
+	    break;
+    if (driver == NULL)
+	return -ENODEV;
+
+    bind_req.dev_info = driver->dev_info;
+    bind_req.Attributes = mtd_info->Attributes;
+    bind_req.Socket = i;
+    bind_req.CardOffset = mtd_info->CardOffset;
+    ret = CardServices(BindMTD, &bind_req);
+    if (ret != CS_SUCCESS) {
+	cs_error(BindMTD, ret);
+	printk("ds: unable to bind MTD '%s' to socket %d offset 0x%lx\n",
+	       (char *)bind_req.dev_info, i, bind_req.CardOffset);
+	return -ENODEV;
+    }
+    return 0;
+} /* bind_mtd */
+
 /*======================================================================
 
     bind_request() connects a socket to a particular client driver.
@@ -761,6 +795,12 @@ static int ds_ioctl(struct inode * inode, struct file * file,
 	ret = CardServices(AccessConfigurationRegister, s->handle,
 			   &buf.conf_reg);
 	break;
+    case DS_GET_FIRST_REGION:
+        ret = CardServices(GetFirstRegion, s->handle, &buf.region);
+	break;
+    case DS_GET_NEXT_REGION:
+	ret = CardServices(GetNextRegion, s->handle, &buf.region);
+	break;
     case DS_BIND_REQUEST:
 	if (!suser()) return -EPERM;
 	err = bind_request(i, &buf.bind_info);
@@ -771,6 +811,12 @@ static int ds_ioctl(struct inode * inode, struct file * file,
     case DS_UNBIND_REQUEST:
 	err = unbind_request(i, &buf.bind_info);
 	break;
+    case DS_BIND_MTD:
+	if (!suser()) return -EPERM;
+	err = bind_mtd(i, &buf.mtd_info);
+	break;
+    default:
+	err = -EINVAL;
     }
     
     if ((err == 0) && (ret != CS_SUCCESS)) {
@@ -812,7 +858,7 @@ int init_module(void)
     servinfo_t serv;
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("ds: Card Services release does not match!");
+	printk("ds: Card Services release does not match!\n");
 	return -1;
     }
     ds_init(0,0);
