@@ -53,7 +53,7 @@ static int major_dev = 0;
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-"pcmem_cs.c 1.39 1995/04/14 19:00:44 (David Hinds)\n";
+"pcmem_cs.c 1.41 1995/05/15 01:43:40 (David Hinds)\n";
 #endif
 
 /*====================================================================*/
@@ -326,6 +326,7 @@ static u_long get_size(dev_link_t *link, pcmem_dev_t *dev)
 
     dev->size = 0;
     mod.Attributes = WIN_ENABLE | WIN_MEMORY_TYPE_CM;
+    mod.AccessSpeed = 0;
     ret = CardServices(ModifyWindow, link->win, &mod);
     if (ret != CS_SUCCESS) {
 	cs_error(ModifyWindow, ret);
@@ -751,6 +752,7 @@ static void do_pcmem_request(void)
 	dev = (pcmem_dev_t *)link->priv;
 	
 	mod.Attributes = WIN_ENABLE;
+	mod.AccessSpeed = 0;
 	ret = CardServices(ModifyWindow, link->win, &mod);
 	if (ret != CS_SUCCESS) {
 	    cs_error(ModifyWindow, ret);
@@ -822,7 +824,7 @@ int init_module(void)
     
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("pcmem_cs: Card Services release does not match!");
+	printk("pcmem_cs: Card Services release does not match!\n");
 	return -1;
     }
     
