@@ -37,7 +37,7 @@
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static const char *version =
-    "tcic.c 1.52 1995/04/11 18:25:08 (David Hinds)\n";
+    "tcic.c 1.53 1995/05/15 01:44:37 (David Hinds)\n";
 #endif
 
 /*====================================================================*/
@@ -855,7 +855,7 @@ int init_module(void)
     servinfo_t serv;
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("tcic: Card Services release does not match!");
+	printk("tcic: Card Services release does not match!\n");
 	return -1;
     }
     return tcic_init();
